//
// Created by yorgo on 07/06/20.
//

#ifndef SVGIMPORT_COMMON_H
#define SVGIMPORT_COMMON_H

#define BOOST_MPL_LIMIT_SET_SIZE 50
#define BOOST_PARAMETER_MAX_ARITY 11

#define SVGPP_USE_EXTERNAL_PATH_DATA_PARSER
#define SVGPP_USE_EXTERNAL_TRANSFORM_PARSER
#define SVGPP_USE_EXTERNAL_PRESERVE_ASPECT_RATIO_PARSER
// #define SVGPP_USE_EXTERNAL_PAINT_PARSER
#define SVGPP_USE_EXTERNAL_MISC_PARSER
// #define SVGPP_USE_EXTERNAL_COLOR_PARSER
#define SVGPP_USE_EXTERNAL_LENGTH_PARSER

#endif //SVGIMPORT_COMMON_H
